/*
 * fdrw.c - disk image <--> file
 * original code:SASANO Takayoshi
 * --- public domain, no warranty.
 */

#include <basic.h>
#include <bstdio.h>
#include <bstdlib.h>
#include <errcode.h>
#include <btron/btron.h>
#include <device/disk.h>

#ifdef DEBUG
#define	DP(x)	printf x
#else
#define	DP(x)	/* */
#endif

LOCAL	TC	target_drive[][4] = {
    {L""}, {L""}, {L""}, {L""},
    {L""}, {L""}, {L""}, {L""},
};


LOCAL WERR	fdw(W argc, TC *argv[])
{
    WERR	err;
    W		dd, fd, imgsize, rsz;
    UB		*buf;
    LINK	link;
    DiskInfo	di;

    /* ǥХ򥪡ץ */
    err = opn_dev(target_drive[*argv[1] - L''], D_UPDATE | D_EXCL, NULL);
    if(err < ER_OK) {
	printf("opn_dev(): %d\n", err);
	goto	fin0;
    }
    dd = err;

    /* ǥΥå */
    err = rea_dev(dd, DN_DISKINFO, (UB *)&di, sizeof(di), &rsz, NULL);
    if(err < ER_OK) {
	printf("rea_dev(): %d\n", err);
	goto	fin1;
    }
    if(di.format != DiskFmt_2DD && di.format != DiskFmt_2HD) {
	printf("unsupported media: %d\n", di.format);
	err = ER_DEV;
	goto fin1;
    }
    DP(("di.format: %d\n", di.format));

    /* ϥեؤΥ󥯼Ф */
    err = get_lnk(argv[2], &link, F_NORM);
    if(err < 0) {
	printf("get_lnk(): %d\n", err);
	goto fin1;
    }

    /* ϥեΥץ */
    err = opn_fil(&link, F_READ | F_EXCL, NULL);
    if(err < 0) {
	printf("opn_fil(): %d\n", err);
	goto fin1;
    }
    fd = err;

    /* ե륵μ */
    err = rea_rec(fd, 0, NULL, 0, &rsz, NULL);
    if(err < 0) {
	printf("rea_rec(): %d\n", err);
	goto fin2;
    }

    /* ᡼η׻: ǥ̤ϰ˼ᡢblockñ̤ڤ夲 */
    imgsize  = (rsz + di.blocksize - 1) / di.blocksize;
    if(imgsize > di.blockcont) imgsize = di.blockcont;
    imgsize *= di.blocksize;
    DP(("di.blocksize: %d  di.blockcont: %d  rsz: %d  imgsize: %d\n",
	di.blocksize, di.blockcont, rsz, imgsize));

    /* γ */
    buf = malloc(imgsize);
    if(buf == NULL) {
	printf("malloc(): NULL\n");
	err = ER_NOMEM;
	goto fin2;
    }
    memset(buf, 0, imgsize);

    /* եɤ߹ */
    err = rea_rec(fd, 0, buf, imgsize, NULL, NULL);
    if(err < 0) {
	printf("rea_rec() %d\n", err);
	goto fin3;
    }

    /* ᡼ν񤭹 */
    err = wri_dev(dd, 0, buf, imgsize / di.blocksize, &rsz, NULL);
    if(err < ER_OK) {
	printf("wri_dev(): %d\n", err);
	goto fin3;
    }

    err = ER_OK;

 fin3:
    free(buf);
 fin2:
    cls_fil(fd);
 fin1:
    cls_dev(dd, 0, NULL);	/* dd, 1, NULL ȥǥȴܤˤʤ */
 fin0:
    return err;
}

LOCAL WERR	fdr(W argc, TC *argv[])
{
    WERR	err;
    W		dd, fd, imgsize, rsz;
    UB		*buf;
    LINK	link;
    DiskInfo	di;

    /* ǥХ򥪡ץ */
    err = opn_dev(target_drive[*argv[1] - L''], D_READ | D_EXCL, NULL);
    if(err < ER_OK) {
	printf("opn_dev(): %d\n", err);
	goto	fin0;
    }
    dd = err;

    /* ǥΥå */
    err = rea_dev(dd, DN_DISKINFO, (UB *)&di, sizeof(di), &rsz, NULL);
    if(err < ER_OK) {
	printf("rea_dev(): %d\n", err);
	goto	fin1;
    }
    if(di.format != DiskFmt_2DD && di.format != DiskFmt_2HD) {
	printf("unsupported media: %d\n", di.format);
	err = ER_DEV;
	goto fin1;
    }
    DP(("di.format: %d\n", di.format));

    /* ᡼η׻ */
    imgsize = di.blocksize * di.blockcont;
    DP(("di.blocksize: %d  di.blockcont: %d  imgsize: %d\n",
	di.blocksize, di.blockcont, imgsize));

    /* γ */
    buf = malloc(imgsize);
    if(buf == NULL) {
	printf("malloc(): NULL\n");
	err = ER_NOMEM;
	goto fin1;
    }

    /* ᡼ɤ߹ */
    err = rea_dev(dd, 0, buf, di.blockcont, &rsz, NULL);
    if(err < ER_OK) {
	printf("rea_dev(): %d\n", err);
	goto fin2;
    }

    /* ϥեؤΥ󥯼Ф */
    err = get_lnk(NULL, &link, F_NORM);
    if(err < 0) {
	printf("get_lnk(): %d\n", err);
	goto fin2;
    }

    /* ϥեκ */
    err = cre_fil(&link, argv[2], NULL, 0, F_FIX);
    if(err < 0) {
	printf("cre_fil(): %d\n", err);
	goto fin2;
    }
    fd = err;

    /* 쥳ɤν񤭹 */
    err = apd_rec(fd, buf, imgsize, 31, 0, 0);
    if(err < 0) {
	printf("apd_rec(): %d\n", err);
	goto fin3;
    }

    err = ER_OK;

 fin3:
    cls_fil(fd);
 fin2:
    free(buf);
 fin1:
    cls_dev(dd, 0, NULL);	/* dd, 1, NULL ȥǥȴܤˤʤ */
 fin0:
    return err;
}

WERR main(W argc, TC *argv[])
{
    /* Ȥɽ */
    if(argc < 3 || *argv[1] < L'' || *argv[1] > L'' ||
       (argv[1][1] != L'' && argv[1][1] != L'')) {
	printf("usage:	fdrw [mode] [filename]\n");
	printf("	mode:	[drive]w - write to disk\n");
	printf("		[drive]r - read from disk\n");
	printf("	drive:	a-d: fda-fdd, e-h: uda-udd\n");
	return ER_PAR;
    }

    switch(argv[1][1]) {
    case	L'':	return fdr(argc, argv);
    case	L'':	return fdw(argc, argv);
    default:		return ER_NOSPT;
    }
}
